"""
Module to read / write wav files using numpy arrays

Functions
---------
`read`: Return the sample rate (in samples/sec) and data from a WAV file.

`write`: Write a numpy array as a WAV file.


This file is a (slightly) modified copy of the scipy/io/wavfile.py from scipy
(https://www.scipy.org/), which is licensed under the following terms:

Copyright (c) 2001-2002 Enthought, Inc.  2003-2019, SciPy Developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import sys
import numpy
import struct
import warnings


__all__ = ["WavFileWarning", "read", "write"]


class WavFileWarning(UserWarning):
    pass


WAVE_FORMAT_PCM = 0x0001
WAVE_FORMAT_IEEE_FLOAT = 0x0003
WAVE_FORMAT_EXTENSIBLE = 0xFFFE
KNOWN_WAVE_FORMATS = (WAVE_FORMAT_PCM, WAVE_FORMAT_IEEE_FLOAT)

# assumes file pointer is immediately
#  after the 'fmt ' id


def _read_fmt_chunk(fid, is_big_endian):
    """
    Returns
    -------
    size : int
        size of format subchunk in bytes (minus 8 for "fmt " and itself)
    format_tag : int
        PCM, float, or compressed format
    channels : int
        number of channels
    fs : int
        sampling frequency in samples per second
    bytes_per_second : int
        overall byte rate for the file
    block_align : int
        bytes per sample, including all channels
    bit_depth : int
        bits per sample
    """
    if is_big_endian:
        fmt = ">"
    else:
        fmt = "<"

    size = res = struct.unpack(fmt + "I", fid.read(4))[0]
    bytes_read = 0

    if size < 16:
        raise ValueError("Binary structure of wave file is not compliant")

    res = struct.unpack(fmt + "HHIIHH", fid.read(16))
    bytes_read += 16

    format_tag, channels, fs, bytes_per_second, block_align, bit_depth = res

    if format_tag == WAVE_FORMAT_EXTENSIBLE and size >= (16 + 2):
        ext_chunk_size = struct.unpack(fmt + "H", fid.read(2))[0]
        bytes_read += 2
        if ext_chunk_size >= 22:
            extensible_chunk_data = fid.read(22)
            bytes_read += 22
            raw_guid = extensible_chunk_data[2 + 4 : 2 + 4 + 16]
            # GUID template {XXXXXXXX-0000-0010-8000-00AA00389B71} (RFC-2361)
            # MS GUID byte order: first three groups are native byte order,
            # rest is Big Endian
            if is_big_endian:
                tail = b"\x00\x00\x00\x10\x80\x00\x00\xAA\x00\x38\x9B\x71"
            else:
                tail = b"\x00\x00\x10\x00\x80\x00\x00\xAA\x00\x38\x9B\x71"
            if raw_guid.endswith(tail):
                format_tag = struct.unpack(fmt + "I", raw_guid[:4])[0]
        else:
            raise ValueError("Binary structure of wave file is not compliant")

    if format_tag not in KNOWN_WAVE_FORMATS:
        raise ValueError("Unknown wave file format")

    # move file pointer to next chunk
    if size > (bytes_read):
        fid.read(size - bytes_read)

    return (size, format_tag, channels, fs, bytes_per_second, block_align, bit_depth)


# assumes file pointer is immediately after the 'data' id
def _read_data_chunk(fid, format_tag, channels, bit_depth, is_big_endian, mmap=False):
    if is_big_endian:
        fmt = ">I"
    else:
        fmt = "<I"

    # Size of the data subchunk in bytes
    size = struct.unpack(fmt, fid.read(4))[0]

    # Number of bytes per sample
    bytes_per_sample = bit_depth // 8
    if bit_depth == 8:
        dtype = "u1"
    else:
        if is_big_endian:
            dtype = ">"
        else:
            dtype = "<"
        if format_tag == WAVE_FORMAT_PCM:
            dtype += "i%d" % bytes_per_sample
        else:
            dtype += "f%d" % bytes_per_sample
    if not mmap:
        data = numpy.frombuffer(fid.read(size), dtype=dtype)
    else:
        start = fid.tell()
        data = numpy.memmap(
            fid, dtype=dtype, mode="c", offset=start, shape=(size // bytes_per_sample,)
        )
        fid.seek(start + size)

    if channels > 1:
        data = data.reshape(-1, channels)
    return data


def _skip_unknown_chunk(fid, is_big_endian):
    if is_big_endian:
        fmt = ">I"
    else:
        fmt = "<I"

    data = fid.read(4)
    # call unpack() and seek() only if we have really read data from file
    # otherwise empty read at the end of the file would trigger
    # unnecessary exception at unpack() call
    # in case data equals somehow to 0, there is no need for seek() anyway
    if data:
        size = struct.unpack(fmt, data)[0]
        fid.seek(size, 1)


def _read_riff_chunk(fid):
    str1 = fid.read(4)  # File signature
    if str1 == b"RIFF":
        is_big_endian = False
        fmt = "<I"
    elif str1 == b"RIFX":
        is_big_endian = True
        fmt = ">I"
    else:
        # There are also .wav files with "FFIR" or "XFIR" signatures?
        raise ValueError("File format {}... not " "understood.".format(repr(str1)))

    # Size of entire file
    file_size = struct.unpack(fmt, fid.read(4))[0] + 8

    str2 = fid.read(4)
    if str2 != b"WAVE":
        raise ValueError("Not a WAV file.")

    return file_size, is_big_endian


def read(filename, mmap=False):
    """
    Open a WAV file

    Return the sample rate (in samples/sec) and data from a WAV file.

    Parameters
    ----------
    filename : string or open file handle
        Input wav file.
    mmap : bool, optional
        Whether to read data as memory-mapped.
        Only to be used on real files (Default: False).

        .. versionadded:: 0.12.0

    Returns
    -------
    rate : int
        Sample rate of wav file.
    data : numpy array
        Data read from wav file.  Data-type is determined from the file;
        see Notes.

    Notes
    -----
    This function cannot read wav files with 24-bit data.

    Common data types: [1]_

    =====================  ===========  ===========  =============
         WAV format            Min          Max       NumPy dtype
    =====================  ===========  ===========  =============
    32-bit floating-point  -1.0         +1.0         float32
    32-bit PCM             -2147483648  +2147483647  int32
    16-bit PCM             -32768       +32767       int16
    8-bit PCM              0            255          uint8
    =====================  ===========  ===========  =============

    Note that 8-bit PCM is unsigned.

    References
    ----------
    .. [1] IBM Corporation and Microsoft Corporation, "Multimedia Programming
       Interface and Data Specifications 1.0", section "Data Format of the
       Samples", August 1991
       http://www.tactilemedia.com/info/MCI_Control_Info.html

    """
    if hasattr(filename, "read"):
        fid = filename
        mmap = False
    else:
        fid = open(filename, "rb")

    try:
        file_size, is_big_endian = _read_riff_chunk(fid)
        fmt_chunk_received = False
        channels = 1
        bit_depth = 8
        format_tag = WAVE_FORMAT_PCM
        while fid.tell() < file_size:
            # read the next chunk
            chunk_id = fid.read(4)

            if not chunk_id:
                raise ValueError("Unexpected end of file.")
            elif len(chunk_id) < 4:
                raise ValueError("Incomplete wav chunk.")

            if chunk_id == b"fmt ":
                fmt_chunk_received = True
                fmt_chunk = _read_fmt_chunk(fid, is_big_endian)
                format_tag, channels, fs = fmt_chunk[1:4]
                bit_depth = fmt_chunk[6]
                if bit_depth not in (8, 16, 32, 64, 96, 128):
                    raise ValueError(
                        "Unsupported bit depth: the wav file "
                        "has {}-bit data.".format(bit_depth)
                    )
            elif chunk_id == b"fact":
                _skip_unknown_chunk(fid, is_big_endian)
            elif chunk_id == b"data":
                if not fmt_chunk_received:
                    raise ValueError("No fmt chunk before data")
                data = _read_data_chunk(
                    fid, format_tag, channels, bit_depth, is_big_endian, mmap
                )
            elif chunk_id == b"LIST":
                # Someday this could be handled properly but for now skip it
                _skip_unknown_chunk(fid, is_big_endian)
            elif chunk_id in (b"JUNK", b"Fake"):
                # Skip alignment chunks without warning
                _skip_unknown_chunk(fid, is_big_endian)
            else:
                warnings.warn(
                    "Chunk (non-data) not understood, skipping it.", WavFileWarning
                )
                _skip_unknown_chunk(fid, is_big_endian)
    finally:
        if not hasattr(filename, "read"):
            fid.close()
        else:
            fid.seek(0)

    return fs, data


def write(filename, rate, data):
    """
    Write a numpy array as a WAV file.

    Parameters
    ----------
    filename : string or open file handle
        Output wav file.
    rate : int
        The sample rate (in samples/sec).
    data : ndarray
        A 1-D or 2-D numpy array of either integer or float data-type.

    Notes
    -----
    * Writes a simple uncompressed WAV file.
    * To write multiple-channels, use a 2-D array of shape
      (Nsamples, Nchannels).
    * The bits-per-sample and PCM/float will be determined by the data-type.

    Common data types: [1]_

    =====================  ===========  ===========  =============
         WAV format            Min          Max       NumPy dtype
    =====================  ===========  ===========  =============
    32-bit floating-point  -1.0         +1.0         float32
    32-bit PCM             -2147483648  +2147483647  int32
    16-bit PCM             -32768       +32767       int16
    8-bit PCM              0            255          uint8
    =====================  ===========  ===========  =============

    Note that 8-bit PCM is unsigned.

    References
    ----------
    .. [1] IBM Corporation and Microsoft Corporation, "Multimedia Programming
       Interface and Data Specifications 1.0", section "Data Format of the
       Samples", August 1991
       http://www.tactilemedia.com/info/MCI_Control_Info.html

    """
    if hasattr(filename, "write"):
        fid = filename
    else:
        fid = open(filename, "wb")

    fs = rate

    try:
        dkind = data.dtype.kind
        if not (
            dkind == "i" or dkind == "f" or (dkind == "u" and data.dtype.itemsize == 1)
        ):
            raise ValueError("Unsupported data type '%s'" % data.dtype)

        header_data = b""

        header_data += b"RIFF"
        header_data += b"\x00\x00\x00\x00"
        header_data += b"WAVE"

        # fmt chunk
        header_data += b"fmt "
        if dkind == "f":
            format_tag = WAVE_FORMAT_IEEE_FLOAT
        else:
            format_tag = WAVE_FORMAT_PCM
        if data.ndim == 1:
            channels = 1
        else:
            channels = data.shape[1]
        bit_depth = data.dtype.itemsize * 8
        bytes_per_second = fs * (bit_depth // 8) * channels
        block_align = channels * (bit_depth // 8)

        fmt_chunk_data = struct.pack(
            "<HHIIHH",
            format_tag,
            channels,
            fs,
            bytes_per_second,
            block_align,
            bit_depth,
        )
        if not (dkind == "i" or dkind == "u"):
            # add cbSize field for non-PCM files
            fmt_chunk_data += b"\x00\x00"

        header_data += struct.pack("<I", len(fmt_chunk_data))
        header_data += fmt_chunk_data

        # fact chunk (non-PCM files)
        if not (dkind == "i" or dkind == "u"):
            header_data += b"fact"
            header_data += struct.pack("<II", 4, data.shape[0])

        # check data size (needs to be immediately before the data chunk)
        if ((len(header_data) - 4 - 4) + (4 + 4 + data.nbytes)) > 0xFFFFFFFF:
            raise ValueError("Data exceeds wave file size limit")

        fid.write(header_data)

        # data chunk
        fid.write(b"data")
        fid.write(struct.pack("<I", data.nbytes))
        if data.dtype.byteorder == ">" or (
            data.dtype.byteorder == "=" and sys.byteorder == "big"
        ):
            data = data.byteswap()
        _array_tofile(fid, data)

        # Determine file size and place it in correct
        #  position at start of the file.
        size = fid.tell()
        fid.seek(4)
        fid.write(struct.pack("<I", size - 8))

    finally:
        if not hasattr(filename, "write"):
            fid.close()
        else:
            fid.seek(0)


if sys.version_info[0] >= 3:

    def _array_tofile(fid, data):
        # ravel gives a c-contiguous buffer
        fid.write(data.ravel().view("b").data)

else:

    def _array_tofile(fid, data):
        fid.write(data.tostring())



# This file is part of lib6300 library for MIT 6.300 (Signal Processing)
# Copyright (c) 2018-2019 by adam hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import wave
import numpy
import tempfile

from . import wavfile

try:
    import pyaudio
except:
    pyaudio = None

_normalize_funcs = {
    "int32": lambda d: d / 2147483648,
    "int16": lambda d: d / 32768,
    "uint8": lambda d: d / 256 - 0.5,
}


def wav_read(fname):
    """
    Read a wave file.  This will always convert to mono.

    Arguments:
      * fname: a string containing a file name of a WAV file.

    Returns a tuple with 2 elements:
      * a 1-dimensional NumPy array with floats in the range [-1, 1]
        representing samples.  the length of this array will be the number of
        samples in the given wave file.
      * an integer containing the sample rate
    """
    # load in data, convert to float
    fs, data = wavfile.read(fname)
    type_ = data.dtype.name
    data = data.astype(float)

    # convert stereo to mono
    if len(data.shape) > 2:
        raise Exception("too many channels!")
    elif len(data.shape) == 2:
        data = data.mean(1)

    # normalize to range [-1, 1]
    try:
        data = _normalize_funcs[type_](data)
    except KeyError:
        pass

    return data.tolist(), fs


def wav_write(samples, fs, fname):
    """
    Write a wave file.

    Arguments:
      * samples: a Python list of numbers in the range [-1, 1], one for each
                 sample in the output WAV file.  Numbers in the list that are
                 outside this range will be clipped to -1 or 1.
      * fs: an integer representing the sampling rate of the output
            (samples/second).
      * fname: a string containing a file name of the WAV file to be written.
    """
    out = numpy.array(samples)
    out[out > 1.0] = 1.0
    out[out < -1.0] = -1.0
    out = (out * 32767).astype("int16")
    wavfile.write(fname, fs, out)


def wav_play(samples, fs):
    """
    Play WAV data from a Python object and a sampling rate.  REQUIRES PYAUDIO.

    Arguments:
      * samples: a Python list of numbers in the range [-1, 1], one for each
                 sample in the output WAV file.  Numbers in the list that are
                 outside this range will be clipped to -1 or 1.
      * fs: an integer representing the sampling rate of the output
            (samples/second).
    """
    if pyaudio is None:
        print(
            "pyaudio is required for playing WAV files directly from lib6300.",
            file=sys.stderr,
            flush=True,
        )
        return
    filename = os.path.join(
        tempfile.gettempdir(), "6300_wave_%s.wav" % abs(hash((fs, tuple(samples))))
    )
    wav_write(samples, fs, filename)
    f = wave.open(filename, "r")
    try:
        p = pyaudio.PyAudio()
        stream = p.open(
            format=p.get_format_from_width(f.getsampwidth()),
            channels=f.getnchannels(),
            rate=f.getframerate(),
            output=True,
        )

        data = f.readframes(10240)
        while data:
            stream.write(data)
            data = f.readframes(10240)

        stream.stop_stream()
        stream.close()
        p.terminate()
    finally:
        f.close()
        os.unlink(filename)


def wav_file_play(fname):
    """
    Play audio from a WAV file on disk using wav_play.  REQUIRES PYAUDIO.

    Arguments:
      * fname: a string containing a file name of the WAV file to be written.
    """
    wav_play(*wav_read(fname))



# This file is part of lib6300, software for use in MIT's 6.300
# Copyright (c) 2018-2019 by the 6.300 Staff <6.300-core@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.

import os

from setuptools import setup

from lib6300 import __version__ as VERSION


def main():
    with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), "r") as f:
        requirements = f.read().split("\n")

    with open(os.path.join(os.path.dirname(__file__), "README"), "r") as f:
        readme = f.read()

    setup(
        name="lib6300",
        version=VERSION,
        author="adam hartz",
        author_email="hz@mit.edu",
        packages=["lib6300"],
        scripts=[],
        url="https://sigproc.mit.edu",
        license="GPLv3+",
        description="Software for 6.300",
        long_description=readme,
        include_package_data=True,
        entry_points={},
        install_requires=requirements,
        package_dir={"lib6300": "lib6300"},
    )


if __name__ == "__main__":
    main()



# This file is part of lib6300 library for MIT 6.300 (Signal Processing)
# Copyright (c) 2018-2019 by adam hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import numpy
import matplotlib.pyplot as plt

from PIL import Image
from math import e, pi
from numpy.fft import fftshift, ifftshift

from .fft import fft2, ifft2


def png_read(fname, zero_loc="center"):
    """
    Load the PNG image with the given filename into a numpy array.  The
    resulting array will have values between 1 (brightest white) and 0 (darkest
    black).

    fname: string
        The filename to load (including .png)

    zero_loc (optional): string
        This parameter controls which point in the original image is considered
        to be (0,0).  By default, it is the center of the given image
        ('center').  The other option is 'topleft'.
    """
    assert zero_loc in {"center", "topleft"}, (
        "zero_loc must be 'center' or 'topleft', not %r" % zero_loc
    )
    assert os.path.isfile(fname), "Not a file: %r" % fname

    with Image.open(fname) as image:
        im_arr = numpy.frombuffer(image.convert("L").tobytes(), dtype=numpy.uint8)
        im_arr = im_arr.reshape((image.size[1], image.size[0]))
    if zero_loc == "center":
        im_arr = ifftshift(im_arr)
    return im_arr.astype(numpy.float64) / 255


def png_write(array, fname, zero_loc="center", normalize=False, zoom=1):
    """
    Save the given numpy array as an image with the given filename.  1
    translates to the brightest white in the output image, and 0 to the darkest
    black.  Values above 1 will be displayed as white, and values below 0 will
    be displayed as black.

    array: numpy.ndarray
        An array containing pixel values

    fname: string
        The name of the file to save (including .png)

    zero_loc (optional): string
        This parameter controls where the point (0,0) should be rendered in the
        output image.  By default, (0,0) is at the center of the image
        ('center').  The other option is 'topleft'.

    normalize (optional): boolean
        If True, normalize values so that the largest element maps to 1
        (brightest white) and the smallest element maps to 0 (darkest black).
        Default: False

    zoom (optional): int
        Factor by which to zoom when saving.
    """
    assert zero_loc in {"center", "topleft"}, (
        "zero_loc must be 'center' or 'topleft', not %r" % zero_loc
    )
    assert (abs(array.imag) <= 1e-6).all(), "input array must contain real values only"
    if zero_loc == "center":
        array = fftshift(array)
    if normalize:
        array = array - numpy.min(array)
        array = array / numpy.max(array)
    h, w = array.shape
    out = Image.new(mode="L", size=(w, h))
    out.putdata(numpy.around(255 * array).reshape(w * h))
    out.resize((int(w * zoom), int(h * zoom)), Image.NEAREST).save(fname)


def show_image(
    array,
    colorscale="linear",
    zero_loc="center",
    cmap="gray",
    show=True,
    title="",
    xlabel="",
    ylabel="",
    vmax=None,
    vmin=None,
):
    """
    Display an image using matplotlib.

    array: numpy.ndarray
        An array containing the image to be displayed

    colorscale(optional): string
        If set to 'log', the colors are plotted on a log scale.  Defaults to
        'linear'.

    zero_loc (optional): string
        This parameter controls where the point (0,0) should be rendered when
        plotting the image.  By default, (0,0) is at the center of the image
        ('center').  The other option is 'topleft'.

    cmap (optional): string
        The color map to use when displaying the image.  Defaults to 'gray'.
        See https://matplotlib.org/examples/color/colormaps_reference.html
        for options.

    show (optional): bool
        If set to True (the default), the plot will be immediately displayed.
        Otherwise, plt.show() will need to be called later to display the plot.

    title (optional): string
        A title for the plot

    xlabel (optional): string
        A label for the horizontal axis

    ylabel (optional): string
        A label ver the vertical axis
    """
    assert colorscale in {"linear", "log"}, (
        "coloscale must be 'linear' or 'log', not %r" % colorscale
    )
    assert zero_loc in {"center", "topleft"}, (
        "zero_loc must be 'center' or 'topleft', not %r" % zero_loc
    )
    assert (abs(array.imag) <= 1e-6).all(), "input array must contain real values only"
    array = array.real
    plt.figure()
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if colorscale == "log":
        array[array != 0] = numpy.log10(array[array != 0])
        array[array == 0] = numpy.min(array - 1)
    h, w = array.shape
    if zero_loc == "center":
        bounds = [
            -w // 2 - 0.5 + (w % 2),
            w // 2 - 0.5 + (w % 2),
            h // 2 - 0.5 + (w % 2),
            -h // 2 - 0.5 + (w % 2),
        ]
        array = fftshift(array)
    else:
        bounds = [0, w, h, 0]
    i = plt.imshow(
        array,
        cmap=cmap,
        interpolation="nearest",
        origin="upper",
        extent=bounds,
        vmin=vmin,
        vmax=vmax,
    )
    plt.colorbar()
    if show:
        plt.show()


def show_dft(array, color_scale="linear", color_limits=None, color_factor=1, show=True):
    """
    Given an image in the spatial domain, plot the magnitude and phase of its
    DFT coefficients using the show_image function.  Magnitude is plotted on a
    log scale.

    array: numpy.ndarray
        An array containing a spatial-domain image (NOT an array of DFT
        coefficients)

    color_scale (optional): string
        Defaults to 'linear'.  If 'log', magnitude colors are on a log scale.

    color_limits (optional): tuple of numbers
        The values of magnitude that should be displayed as black and white,
        respectively.  Values outside this range will be clipped to white or
        black.

    color_factor (optional): number
        A value by which the magnitude should be scaled for display purposes.

    show (optional): bool
        If set to True (the default), the plot will be immediately displayed.
        Otherwise, plt.show() will need to be called later to display the plot.
    """
    X = fft2(array)
    phase = numpy.angle(X)
    mag = abs(X)
    phase[mag <= 1e-6] = 0
    if color_limits is None:
        if color_scale == "log":
            color_limits = (None, None)
        else:
            color_limits = (0, 1)
    show_image(
        mag * color_factor,
        colorscale=color_scale,
        show=False,
        title=r"$\left|X[k_x, k_y]\right|$",
        xlabel="$k_x$",
        ylabel="$k_y$",
        vmin=color_limits[0],
        vmax=color_limits[1],
    )
    show_image(
        phase,
        cmap="coolwarm",
        title=r"$\angle X[k_x, k_y]$",
        xlabel="$k_x$",
        ylabel="$k_y$",
        vmax=pi,
        vmin=-pi,
        show=show,
    )



# This file is part of lib6300 library for MIT 6.300 (Signal Processing)
# Copyright (c) 2018-2019 by adam hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy
from numpy.fft import fft as nfft, ifft as nifft, fft2 as nfft2, ifft2 as nifft2


def fft(x):
    """
    fft function that is simply a wrapper around numpy's (very fast) FFT,
    modified to correct the scaling factor to be consistent with 6.300's
    definitions.
    """
    if isinstance(x, list):
        x = numpy.array(x)
        islist = True
    else:
        islist = False
    out = nfft(x) / (x.size)
    return out.tolist() if islist else out


def ifft(x):
    """
    Similarly, a wrapper around numpy's IFFT to correct the scaling factor to
    be consistent with 6.300's definitions.
    """
    if isinstance(x, list):
        x = numpy.array(x)
        islist = True
    else:
        islist = False
    out = nifft(x) * (x.size)
    return out.tolist() if islist else out


def fft2(x):
    """
    fft function that is simply a wrapper around numpy's (very fast) 2-D FFT,
    modified to correct the scaling factor to be consistent with 6.300's
    definitions.
    """
    x = numpy.array(x)
    return nfft2(x) / (x.size)


def ifft2(x):
    """
    Similarly, a wrapper around numpy's 2-D IFFT to correct the scaling factor to
    be consistent with 6.300's definitions.
    """
    x = numpy.array(x)
    return nifft2(x) * (x.size)
